/*
 * SilberLand MUD-Lib
 * ---
 * /secure/npcmaster.c
 * NPC Score Master, maintained by Woody
 * ---
 * $Date: 1997/06/28 20:03:02 $
 * $Revision: 1.5 $
 * $Log: npcmaster.c,v $
 * Revision 1.5  1997/06/28 20:03:02  Woody
 * Bugfix, wenn this_player() in QueryNPC() nicht definiert war
 * (wie zum Beispiel beim Bumerang -> dann P_KILLER verwenden)
 *
 * Revision 1.4  1997/06/08 02:01:33  Woody
 * Bugfix in QueryNPC()
 *
 * Revision 1.3  1997/06/01 15:56:41  Woody
 * Queue Handling
 *
 * Revision 1.2  1997/03/15 23:27:11  Woody
 * prepared for revision control
 *
 * Revision 1.1  96-11-13  Woody
 * Abfrage, ob Testspieler, oder Filename /test/ enthaelt
 *
 * Revision 1.0  96-10-05  Woody
 * Erste Implementation - uebernommen von der MG-mudlib
 */

#include <npcmaster.h>
#include <wizlevels.h>
#include <properties.h>

// prototypes

varargs mixed QueryNPC(int score);
int SetScore(mixed key, int score);
mixed DelLast();
static int DumpNPCs();
mixed *QueryNPCbyNumber(int num);
mixed RemoveQueued(string key);
mixed RemoveQueuedRegexp(string pattern);
mixed ApproveQueued(string key);
mixed ApproveQueuedRegexp(string pattern);
varargs mapping QueryQueued(mixed pattern);
mapping QueryPending();
int Recalculate(string s);

// data

private int lastNum;
private mapping npcs;                   // NPC Mapping
private mapping queued;                 // noch zu bestaetigende Scores
private mapping pending;                // noch zu vergebende Scores
private static mapping by_num;          // Hilfsmapping (by number)

void make_num(string key)
{
  by_num += ([ npcs[key,NPC_NUMBER] : key; npcs[key,NPC_SCORE] ]);
}

void create()
{
  seteuid(getuid());
  npcs=allocate_mapping(0,2);
  queued=allocate_mapping(0,1);
  pending=allocate_mapping(0,1);
  if (!restore_object(NPCSAVEFILE))
  {
    lastNum=0;
    npcs=allocate_mapping(0,2);
    queued=allocate_mapping(0,1);
    pending=allocate_mapping(0,1);
  }
  by_num = allocate_mapping(0,2);
  filter_mapping(npcs, #'make_num/*'*/);
}

void reset()
{
  string who;
  mixed u,ind;
  int i,j,changed;

  u=users();
  for (j=sizeof(u)-1; j>=0; j--)
  {
    who=getuid(u[j]);
    if (pointerp((ind=pending[who])))
    {
      for (i=sizeof(ind)-1; i>=0; i--)
      {
        if (member(npcs,ind[i])) {
          u[j]->AddScore(npcs[ind[i],NPC_NUMBER]);
          pending[who] -= ({ ind[i] });
          changed=1;
        }
        else if (!member(queued,ind[i])) {
          pending[who] -= ({ ind[i] });
          changed=1;
        }
      }
      if (!sizeof(pending[who])) pending=m_delete(pending,who);
    }
  }
  if (changed) save_object(NPCSAVEFILE);
}

varargs mixed QueryNPC(int score)
{
  string key, who;
  object ob,killer;
  mixed ind;
  int i;

  if (!(ob=previous_object()))
    return NPC_INVALID_ARG;

  killer=this_player();
  if (!killer) killer=ob->QueryProp(P_KILLER);
  if (!killer || !interactive(killer)) return NPC_INVALID_ARG;

  // FIRST OF ALL process pending scores!
  who=getuid(killer);
  if (pointerp((ind=pending[who])))
  {
    for (i=sizeof(ind)-1; i>=0; i--)
    {
      if (member(npcs,ind[i])) {
        killer->AddScore(npcs[ind[i],NPC_NUMBER]);
        pending[who] -= ({ ind[i] });
      }
      else if (!member(queued,ind[i]))
        pending[who] -= ({ ind[i] });
    }
    if (!sizeof(pending[who])) pending=m_delete(pending,who);
  }

  key=explode(file_name(previous_object()),"#")[0];
  if (member(npcs,key))
    return ({npcs[key,NPC_NUMBER],npcs[key,NPC_SCORE]});

  if (killer->QueryProp(P_TESTPLAYER) || IS_WIZARD(killer))
  {
    pending=m_delete(pending,who);      // evtl. aus mapping raus
    return NPC_INVALID_ARG;
  }
  if (strstr(key, "/test/") != -1)
    return NPC_INVALID_ARG;
  if (score<=0 ||
      member(inherit_list(previous_object()),"std/living/life.c") < 0)
    return NPC_INVALID_ARG;

  queued[key]=score;
  if (member(pending,who)) {
    if (member(pending[who],key) == -1)
      pending[who] += ({ key });
  } else
    pending[who] = ({ key });

  save_object(NPCSAVEFILE);
  return NPC_QUEUED;

  /*
  npcs[key,NPC_SCORE]=score;
  npcs[key,NPC_NUMBER]=++lastNum;
  by_num += ([lastNum:key;score]);
  save_object(NPCSAVEFILE);
  call_out("DumpNPCs",0);
  return ({lastNum,score});
  */
}

int SetScore(mixed key, int score)
{
  int x;
  if (!IS_ARCH(this_interactive())||!IS_ARCH(geteuid(previous_object())))
    return NPC_NO_PERMISSION;
  if (intp(key))
  {
    string *ind;
    ind=m_indices(npcs);
    for (x=sizeof(npcs)-1; x>=0; x--)
      if (npcs[ind[x],NPC_NUMBER] == key)
      {
        key=ind[x];
        break;
      }
      if (intp(key)) return NPC_INVALID_ARG;
  }
  if (!stringp(key)) return NPC_INVALID_ARG;
  if (!member(npcs,key))
    npcs[key,NPC_NUMBER]=++lastNum;
  npcs[key,NPC_SCORE]=score;
  by_num += ([npcs[key,NPC_NUMBER]:key;score]);
  queued = m_delete(queued,key);
  save_object(NPCSAVEFILE);
  DumpNPCs();
  return 1;
}

mixed DelLast()
{
  string s;
  if (!IS_ARCH(this_interactive())||!IS_ARCH(geteuid(previous_object())))
    return NPC_NO_PERMISSION;
  s="Deleted #"+lastNum+" ["+by_num[lastNum]+"]\n";
  npcs = m_delete (npcs, by_num[lastNum]);
  by_num = m_delete (by_num, lastNum);
  lastNum--;
  save_object(NPCSAVEFILE);
  call_out("DumpNPCs",0);
  return s;
}

static int DumpNPCs()
{
  mixed *keys;
  int i, max;

  while (remove_call_out("DumpNPCs") != -1) /* continue */ ;
  keys=m_indices(npcs);
  keys=sort_array(keys, #'<);
  rm(NPCDUMPFILE);
  for (i=sizeof(keys)-1, max=0;i>=0;i--) {
    write_file(NPCDUMPFILE,sprintf("%5d %4d %s\n",npcs[keys[i],NPC_NUMBER],
	   npcs[keys[i],NPC_SCORE],keys[i]));
    max += npcs[keys[i],NPC_SCORE];
  } 
  write_file(NPCDUMPFILE,sprintf("====================\n"+
             "Gesamt: %d Punkte\n",max)); 
  return 1;
}

mixed *QueryNPCbyNumber(int num)
{
  if (by_num[num])
    return ({num,by_num[num,NPC_SCORE],by_num[num,NPC_NUMBER]});

  return 0;
}

mixed RemoveQueued(string key)
{
  if (!IS_ARCH(this_interactive())||!IS_ARCH(geteuid(previous_object())))
    return NPC_NO_PERMISSION;
  queued = m_delete(queued,key);
  save_object(NPCSAVEFILE);
  return copy_mapping(queued);
}

mixed RemoveQueuedRegexp(string pattern)
{
  int i;
  string *match;
  if (!IS_ARCH(this_interactive())||!IS_ARCH(geteuid(previous_object())))
    return NPC_NO_PERMISSION;
  match=regexp(m_indices(queued),pattern);
  for (i=sizeof(match)-1; i>=0; i--)
    queued = m_delete(queued,match[i]);
  save_object(NPCSAVEFILE);
  return copy_mapping(queued);
}

mixed ApproveQueued(string key)
{
  if (!IS_ARCH(this_interactive())||!IS_ARCH(geteuid(previous_object())))
    return NPC_NO_PERMISSION;
  if (!stringp(key)) return 0;
  if (member(queued,key) && !member(npcs,key)) {
    npcs[key,NPC_SCORE]=queued[key];
    npcs[key,NPC_NUMBER]=++lastNum;
    by_num += ([lastNum:key;queued[key]]);
    queued = m_delete(queued,key);
    save_object(NPCSAVEFILE);
    call_out("DumpNPCs",0);
    return ({npcs[key,NPC_NUMBER],npcs[key,NPC_SCORE]});
  }
  else return NPC_INVALID_ARG;
}

mixed ApproveQueuedRegexp(string pattern)
{
  int i;
  string *match;
  mixed added;

  if (!IS_ARCH(this_interactive())||!IS_ARCH(geteuid(previous_object())))
    return NPC_NO_PERMISSION;
  if (!stringp(pattern)) return 0;
  match=regexp(m_indices(queued),pattern);
  added=({});
  for (i=sizeof(match)-1; i>=0; i--)
  {
    if (!member(npcs,match[i])) {
      npcs[match[i],NPC_SCORE]=queued[match[i]];
      npcs[match[i],NPC_NUMBER]=++lastNum;
      by_num += ([lastNum:match[i];queued[match[i]]]);
      queued = m_delete(queued,match[i]);
      added += ({ ({ match[i], npcs[match[i],NPC_NUMBER],
                               npcs[match[i],NPC_SCORE] }) });
    }
  }
  if (!sizeof(added)) return NPC_INVALID_ARG;
  save_object(NPCSAVEFILE);
  call_out("DumpNPCs",0);
  return added;
}

varargs mapping QueryQueued(mixed pattern)
{
  string *match;
  mapping m;
  int i;

  if (!stringp(pattern)) return copy_mapping(queued);
  m=([]);
  match=regexp(m_indices(queued),pattern);
  for (i=sizeof(match)-1; i>=0; i--)
    m[match[i]]=queued[match[i]];
  return m;
}

mapping QueryPending()
{
  return copy_mapping(pending);
}

int Recalculate(string s)
{
  int i, j;

  if (!s || !strlen(s))
    return 0;

  for (j=0, i=6*strlen(s)-1; i>0; i--) {
    if (test_bit(s,i))
      j+=by_num[i,NPC_SCORE];
  }
  return j;
}

