/* Ein Standardbehaelter, aus den man mit kleinen Aenderungen schnell
 * zu einen Lederbeutel, eine Tasche oder einen Sack zu machen. Er ist nicht
 * abschliessbar.
 *
 * Fuer abschliessbare Objekte (magisch, mit Schluessel oder was) bitte
 * 'std/truhe' benutzen! Dieses Objekt kann um einiges mehr.
 *
 * Dokumentation siehe 'man behaelter' bzw 'man truhe'
 *
 * Wed Apr 21 22:12:15 1999, Shannara@Turithil
 *
 * Tue Jul 27 11:57:44 1999, Shannara@Turithil
 * bugfix
 */
#include <properties.h>
#include <container.h>
#include <defines.h>
inherit "std/container";

int	opened;

/* mit dieser Funktion kann man ueberpruefen, ob das Objekt ein Behaelter ist
 * oder etwa nicht
 */
int isBehaelter() { return 1; }

/* QueryAllowed() uebergibt 1, wenn man diesen Container oeffnen UND
 * schliessen darf. Bei Bedarf ueberschreiben.
 */
int QueryAllowed() { return 1; }

/* QueryStatus() gibt den Status (0 == zu, 1 == offen) zurueck */
int QueryStatus() { return opened; }

/* Oeffnen und Schliessen der Truhe, gibt dann Zustand zurueck,
 * beruecksichtigt automatisch QueryAllowed().
 */ 
int Open() 
{ 
    if(!opened && QueryAllowed())
    {
	opened=1;
	SetProp(P_TRANSPARENT, 1);	
    }
    
    return QueryStatus();
}

int Close() 
{
    if(opened && QueryAllowed())
    {
	opened=0;
	SetProp(P_TRANSPARENT, 0);
    }
 
    return QueryStatus();
}

/* ------------------------------------------------------------------------ */
int open_fun(string str)
{
    if(!id(str))
    {
    	notify_fail("Was moechtest du denn oeffnen?\n");
	return 0;
    }
    
    if(QueryStatus() == CNT_OPENED)
    {
	notify_fail(capitalize(name(WER, 1))+" ist bereits offen!\n");
	return 0;
    }
    
    if(Open() == CNT_OPENED)
    {
	write("Du oeffnest "+name(WEN)+".\n");
	say(capitalize(this_player()->name())+" oeffnet "+name(WEN)+".\n");
    }
    else
    {
	write("Irgendwie schaffst du es nicht, "+name(WEN)+" zu oeffnen.\n");
    }    
    return 1;
}

int close_fun(string str)
{
    if(!id(str)) 
    {
	notify_fail("Was moechtest du schliessen?\n");
	return 0;
    }
    
    if(QueryStatus() == CNT_CLOSED)
    {
	notify_fail(capitalize(name(WER, 1))+" ist schon zu!\n");
	return 0;
    }
    
    if(Close() == CNT_CLOSED)
    {
	write("Du schliesst "+name(WEN)+".\n");
	say(capitalize(this_player()->name())+" schliesst "+name(WEN)+".\n");
    }
    else
    {
	write("Irgendwie schaffst du es nicht, "+name(WEN)+" zu schliessen.\n");
    }    
    return 1;
}

/* ------------------------------------------------------------------------ */
string short()
{
    string	sh;
    
    if(QueryProp(P_INVIS)) { return 0; }
    if(sh=QueryProp(P_SHORT))
    {
	if(opened)
	{   return sh+" (geoeffnet).\n"; }
	else 
	{   return sh+" (geschlossen).\n"; }
    }
}

string long()
{
    string	lo, inv;
    
    lo=QueryProp(P_LONG);
    if(!QueryProp(P_TRANSPARENT))
    {
	return lo+capitalize(name(WER))+" ist geschlossen.\n";
    }
    
    inv=make_invlist(this_player(), all_inventory(this_object()));

    if(inv && (inv != ""))
    {	return lo+capitalize(name(WER))+" ist offen und enthaelt:\n"+inv; }
    else
    {	return lo+capitalize(name(WER))+" ist offen und leer.\n"; }
}

init()
{
    ::init();
    
    add_action("open_fun", "oeffne");
    add_action("close_fun", "schliess");
    add_action("close_fun", "schliesse");
}

create()
{
    if(!is_clone(ME)) return 0;
    ::create();
    
    AddId("Behaelter");			/* analog zu Npc, Living, ... */
    AddId("behaelter");
    
    SetProp(P_GENDER, MALE);
    SetProp(P_NAME, "Behaelter");
    SetProp(P_SHORT, "Ein Behaelter");
    SetProp(P_LONG, "Ein einfacher Behaelter ohne Eigenschaften.\n");
    SetProp(P_TRANSPARENT, 0);
    SetProp(P_MAX_WEIGHT, 0);		/* wieviel Gramm passen rein */
    SetProp(P_WEIGHT_PERCENT, 0);	/* wieviel % kommt nach aussen */
    
    opened=0;
}